@@ -5,6 +5,11 @@ import uuid
 import copy
 from vipy.object import Track
 
+try:
+    import ujson as json  # faster
+except ImportError:
+    import json
+
 
 class Activity(object):
     """vipy.object.Activity class
@@ -30,18 +35,36 @@ class Activity(object):
             if all([isinstance(t, Track) for t in tracks]):
                 assert all([any([t.during(f) for f in range(startframe, endframe)]) for t in tracks]), "All tracks must be be present in at least one frame when this activity occurs"
                 tracks = [t.id() for t in tracks]  
-               
-        self._id = uuid.uuid1().hex
-        self._startframe = startframe
-        self._endframe = endframe
+        trackid = set(tracks) if tracks is not None else set([])  # only store IDs, not tracks
+        
+        if actorid is not None:
+            assert isstring(actorid), "Invalid actor ID, must be a track ID for the actor performing this activity"
+            if actorid not in trackid:
+                trackid.add(actorid)
+
+        self._id = uuid.uuid4().hex
+        self._startframe = int(startframe)
+        self._endframe = int(endframe)
         self._framerate = framerate
         self._label = category if category is not None else label        
         self._shortlabel = self._label if shortlabel is None else shortlabel
-        self._trackid = set(tracks) if tracks is not None else set([])
+        self._trackid = trackid
         self._actorid = actorid
 
         self.attributes = attributes if attributes is not None else {}            
-        
+
+    @classmethod
+    def from_json(obj, s):
+        d = json.loads(s) if not isinstance(s, dict) else s                
+        return obj(startframe=d['_startframe'],
+                   endframe=d['_endframe'],
+                   framerate=d['_framerate'],
+                   category=d['_label'],
+                   shortlabel=d['_shortlabel'],
+                   tracks=d['_trackid'],
+                   attributes=d['attributes'],
+                   actorid=d['_actorid'])
+                
     def __len__(self):
         """Return activity length in frames, or zero if degenerate"""
         return max(0, self.endframe() - self.startframe())
@@ -50,8 +73,12 @@ class Activity(object):
         return str('<vipy.activity: category="%s", frames=(%d,%d), tracks=%s>' % (self.category(), self.startframe(), self.endframe(), len(self.trackids())))
 
     def dict(self):
-        return {'id':self._id, 'label':self.category(), 'shortlabel':self.shortlabel(), 'startframe':self._startframe, 'endframe':self._endframe, 'attributes':self.attributes, 'framerate':self._framerate,
-                'trackid':self._trackid, 'actorid':self._actorid}
+        """Return a python dictionary containing the relevant serialized attributes suitable for JSON encoding"""
+        return self.json(s=None, encode=False)
+
+    def json(self, encode=True):
+        d = {k:v if k != '_trackid' else list(v) for (k,v) in self.__dict__.items()}
+        return json.dumps(d) if encode else d
     
     def actorid(self, actorid=None):
         if actorid is None:
@@ -71,7 +98,7 @@ class Activity(object):
         if f is None:
             return self._endframe
         else:
-            self._endframe = f
+            self._endframe = int(f)
             return self
 
     def middleframe(self):
@@ -154,6 +181,10 @@ class Activity(object):
         """Is frame during the time interval (startframe, endframe) inclusive?"""
         return int(frame) >= self._startframe and int(frame) <= self._endframe
 
+    def during_interval(self, startframe, endframe):
+        """Is the activity occurring for all frames within the interval [startframe, endframe) (non-inclusive of endframe)?"""
+        return all([self.during(f) for f in range(startframe, endframe)])
+    
     def temporal_iou(self, other):
         """Return the temporal intersection over union of two activities"""
         assert isinstance(other, Activity), "Invalid input - must be vipy.object.Activity()"
@@ -168,9 +199,15 @@ class Activity(object):
         return (t_intersection / t_union) if t_intersection > 0 else 0
     
     def offset(self, dt):
-        self._startframe = self._startframe + dt
-        self._endframe = self._endframe + dt
+        self._startframe = int(self._startframe + dt)
+        self._endframe = int(self._endframe + dt)
         return self
+
+    def truncate(self, startframe=None, endframe=None):
+        self._startframe = self._startframe if startframe is None else max(self._startframe, startframe)
+        self._endframe = self._endframe if endframe is None else max(self._endframe, endframe)
+        return self
+        
     
     def id(self, newid=None):
         if newid is None:
