@@ -3,6 +3,7 @@ import matplotlib
 from matplotlib import pyplot as plt
 import numpy as np
 from vipy.util import islist, temppng
+import sys
 
 
 FIGHANDLE = {}
@@ -11,7 +12,8 @@ matplotlib.rcParams['toolbar'] = 'None'
 
 # Optional latex strings in captions
 try:
-    if 'VIPY_LATEX' in os.environ:
+    import vipy.globals
+    if vipy.globals.GLOBAL['LATEX'] is not None:
         from distutils.spawn import find_executable
         if not find_executable('latex'):
             raise
@@ -20,18 +22,33 @@ except:
     pass  # ignored if latex is not installed or not wanted
 
 
+def escape_to_exit(event):
+    if event.key == 'escape' or event.key == 'q' or event.key == 'ctrl+c':
+        import vipy.globals            
+        vipy.globals.user_hit_escape(True)
+    
 def flush():
     plt.pause(0.001)
 
     
+def imflush():
+    plt.pause(0.001)
+
+    # this is necessary for imshow only, maybe to trigger remove() of child?
+    # However, this breaks jupyter notebook image show if we use gca()
+    [a.annotate('', (0,0)) for i in plt.get_fignums() for a in plt.figure(i).axes]
+    
+    #plt.gca().annotate('', (0,0))  
+
+    
 def show(fignum):
+    fig = plt.figure(fignum) 
+    fig.canvas.draw()    
     plt.ion()
-    plt.draw()
     plt.show()
 
-
+    
 def noshow(fignum):
-    closeall()
     plt.ioff()
 
 
@@ -55,8 +72,9 @@ def figure(fignum=None):
 def close(fignum):
     global FIGHANDLE
     if fignum in FIGHANDLE:
+        plt.close(fignum)
         FIGHANDLE.pop(fignum, None)
-        return plt.close(fignum)
+        return None
     else:
         return None
 
@@ -80,15 +98,23 @@ def _imshow_tight(img, fignum=None):
         a.get_xaxis().set_visible(False)
         a.get_yaxis().set_visible(False)
     imh = plt.imshow(img, animated=True, interpolation='nearest', aspect='equal')
+
+    fig.canvas.mpl_connect('key_press_event', escape_to_exit)
+    import vipy.globals    
+    vipy.globals.user_hit_escape(False)
+
     return (fig.number, imh)
 
 
 def imshow(img, fignum=None):
     """Show an image in a figure window (optionally visible), reuse previous figure if it is the same shape"""
     global FIGHANDLE
+
     if fignum in plt.get_fignums() and fignum in FIGHANDLE and FIGHANDLE[fignum].get_size() == img.shape[0:2]:
-        # Delete all polygon and text overlays from previous drawing
+        # Do not delete and recreate the figure, just change the pixels 
         FIGHANDLE[fignum].set_data(img)
+
+        # Delete all polygon and text overlays from previous drawing so that they can be overwritten on current frame
         for c in plt.gca().get_children():
             if 'Text' in c.__repr__() or 'Polygon' in c.__repr__() or 'Circle' in c.__repr__() or 'Line' in c.__repr__() or 'Patch' in c.__repr__():
                 try:
@@ -96,6 +122,7 @@ def imshow(img, fignum=None):
                 except:
                     pass
     else:
+        # Jupyter notebook does not respect fignum.  It is always one when inspecting plt.get_fignums()
         if fignum in plt.get_fignums() and fignum in FIGHANDLE:
             close(fignum)
         (fignum, imh) = _imshow_tight(img, fignum=fignum)
@@ -103,6 +130,20 @@ def imshow(img, fignum=None):
     return fignum
 
 
+def text(caption, xmin, ymin, fignum=None, textcolor='black', textfacecolor='white', textfacealpha=1.0, fontsize=10, linewidth=3, facecolor='white', facealpha=0.5, alpha=1.0):
+    plt.figure(fignum) if fignum is not None else plt.gcf()
+    lw = linewidth  # pull in the boxes by linewidth so that they do not overhang the figure
+
+    newlines = caption.count('\n')
+    captionoffset = 15*newlines   # move down a bit if near top of image, shift once per newline in caption    
+    try:
+        # MatplotlibDeprecationWarning: The 's' parameter of annotate() has been renamed 'text' since Matplotlib 3.3            
+        handle = plt.annotate(alpha=alpha, text=caption, xy=(xmin,ymin), xytext=(xmin, ymin+captionoffset), xycoords='data', color=textcolor, bbox=None if textfacecolor is None else dict(facecolor=textfacecolor, edgecolor=None, alpha=textfacealpha, boxstyle='square'), fontsize=fontsize, clip_on=True)
+    except:
+        handle = plt.annotate(alpha=alpha, s=caption, xy=(xmin,ymin), xytext=(xmin, ymin+captionoffset), xycoords='data', color=textcolor, bbox=None if textfacecolor is None else dict(facecolor=textfacecolor, edgecolor=None, alpha=textfacealpha, boxstyle='square'), fontsize=fontsize, clip_on=True)        
+
+    return fignum
+
 def boundingbox(img, xmin, ymin, xmax, ymax, bboxcaption=None, fignum=None, bboxcolor='green', facecolor='white', facealpha=0.5, textcolor='black', textfacecolor='white', textfacealpha=1.0, fontsize=10, captionoffset=(0,0), linewidth=3):
     """Draw a captioned bounding box on a previously shown image"""
     plt.figure(fignum)
@@ -117,10 +158,10 @@ def boundingbox(img, xmin, ymin, xmax, ymax, bboxcaption=None, fignum=None, bbox
         newlines = bboxcaption.count('\n')
         captionoffset = captionoffset if ymin > 15 else (captionoffset[0]+5, captionoffset[1]+(15*(newlines+1)))  # move down a bit if near top of image, shift once per newline in caption
         try:
-            handle = plt.annotate(s=bboxcaption, xy=(xmin,ymin), xytext=(xmin+captionoffset[0], ymin+captionoffset[1]), xycoords='data', color=textcolor, bbox=dict(facecolor=textfacecolor, edgecolor=textcolor, alpha=textfacealpha, boxstyle='round'), fontsize=fontsize, clip_on=True)
+            # MatplotlibDeprecationWarning: The 's' parameter of annotate() has been renamed 'text' since Matplotlib 3.3            
+            handle = plt.annotate(text=bboxcaption, xy=(xmin,ymin), xytext=(xmin+captionoffset[0], ymin+captionoffset[1]), xycoords='data', color=textcolor, bbox=dict(facecolor=textfacecolor, edgecolor=textcolor, alpha=textfacealpha, boxstyle='round'), fontsize=fontsize, clip_on=True)
         except:
-            # "text" is the new style
-            handle = plt.annotate(text=bboxcaption, xy=(xmin,ymin), xytext=(xmin+captionoffset[0], ymin+captionoffset[1]), xycoords='data', color=textcolor, bbox=dict(facecolor=textfacecolor, edgecolor=textcolor, alpha=textfacealpha, boxstyle='round'), fontsize=fontsize, clip_on=True)            
+            handle = plt.annotate(s=bboxcaption, xy=(xmin,ymin), xytext=(xmin+captionoffset[0], ymin+captionoffset[1]), xycoords='data', color=textcolor, bbox=dict(facecolor=textfacecolor, edgecolor=textcolor, alpha=textfacealpha, boxstyle='round'), fontsize=fontsize, clip_on=True)            
 
     return fignum
 
@@ -129,8 +170,10 @@ def imdetection(img, detlist, fignum=None, bboxcolor='green', do_caption=True, f
     """Show bounding boxes from a list of vipy.object.Detections on the same image, plotted in list order with optional captions """
 
     # Create image
-    fignum = imshow(img, fignum=fignum)
+    fignum = imshow(img, fignum=fignum) 
 
+    # A better way? https://matplotlib.org/api/_as_gen/matplotlib.animation.FuncAnimation.html
+    
     # Valid detections
     for (k,det) in enumerate(detlist):
         if do_caption:
