@@ -5,6 +5,11 @@ import uuid
 import copy
 import warnings
 
+try:
+    import ujson as json  # faster
+except ImportError:
+    import json
+
 
 class Detection(BoundingBox):
     """vipy.object.Detection class
@@ -19,24 +24,43 @@ class Detection(BoundingBox):
 
     """
 
-    def __init__(self, label=None, xmin=None, ymin=None, width=None, height=None, xmax=None, ymax=None, confidence=None, xcentroid=None, ycentroid=None, category=None, xywh=None, shortlabel=None, attributes=None):
-        super(Detection, self).__init__(xmin=xmin, ymin=ymin, width=width, height=height, xmax=xmax, ymax=ymax, xcentroid=xcentroid, ycentroid=ycentroid, xywh=xywh)
+    def __init__(self, label=None, xmin=None, ymin=None, width=None, height=None, xmax=None, ymax=None, confidence=None, xcentroid=None, ycentroid=None, category=None, xywh=None, shortlabel=None, attributes=None, id=None):
+        super().__init__(xmin=xmin, ymin=ymin, width=width, height=height, xmax=xmax, ymax=ymax, xcentroid=xcentroid, ycentroid=ycentroid, xywh=xywh)
         assert not (label is not None and category is not None), "Constructor requires either label or category kwargs, not both"
-        self._id = uuid.uuid1().hex        
+        self._id = uuid.uuid4().hex if id is None else id
         self._label = category if category is not None else label
         self._shortlabel = self._label if shortlabel is None else shortlabel
         self._confidence = float(confidence) if confidence is not None else confidence
         self.attributes = attributes if attributes is not None else {}
 
+    @classmethod
+    def cast(cls, d, flush=False):
+        assert isinstance(d, BoundingBox)
+        if d.__class__ != Detection:
+            d.__class__ = Detection
+            d._id = uuid.uuid4().hex if flush or not hasattr(d, '_id') else d._id
+            d._shortlabel = None if flush or not hasattr(d, '_shortlabel') else d._shortlabel
+            d._confidence = None if flush or not hasattr(d, '_confidence') else d._confidence
+            d._label = None if flush or not hasattr(d, '_label') else d._label
+            d.attributes = {} if flush or not hasattr(d, 'attributes') else d.attributes
+        return d
+        
+    @classmethod
+    def from_json(obj, s):
+        d = json.loads(s) if not isinstance(s, dict) else s        
+        return obj(xmin=d['_xmin'], ymin=d['_ymin'], xmax=d['_xmax'], ymax=d['_ymax'], label=d['_label'], shortlabel=d['_shortlabel'], confidence=d['_confidence'], attributes=d['attributes'])
+        
     def __repr__(self):
         strlist = []
         if self.category() is not None:
             strlist.append('category="%s"' % self.category())
-        if self.isvalid():
+        if True:
             strlist.append('bbox=(xmin=%1.1f, ymin=%1.1f, width=%1.1f, height=%1.1f)' %
                            (self.xmin(), self.ymin(),self.width(), self.height()))
         if self._confidence is not None:
             strlist.append('conf=%1.3f' % self.confidence())
+        if self.isdegenerate():
+            strlist.append('degenerate')
         return str('<vipy.object.detection: %s>' % (', '.join(strlist)))
 
     def __eq__(self, other):
@@ -47,20 +71,27 @@ class Detection(BoundingBox):
         return self.__repr__()
 
     def dict(self):
-        return {'id':self._id, 'label':self.category(), 'shortlabel':self.shortlabel() ,'boundingbox':super(Detection, self).dict(),
-                'attributes':self.attributes,  # these may be arbitrary user defined objects
-                'confidence':self._confidence}
-                    
+        """Return a python dictionary containing the relevant serialized attributes suitable for JSON encoding"""
+        return self.json(s=None, encode=False)
+
+    def json(self, encode=True):
+        return json.dumps(self.__dict__) if encode else self.__dict__
+                
     def nocategory(self):
         self._label = None
         return self
 
-    def category(self, category=None):
-        """Update the category of the detection"""
+    def noshortlabel(self):
+        self._shortlabel = None
+        return self
+        
+    def category(self, category=None, shortlabel=True):
+        """Update the category and shortlabel (optional) of the detection"""
         if category is None:
             return self._label
         else:
             self._label = str(category)  # coerce to string
+            self._shortlabel = str(category) if shortlabel else self._shortlabel  # coerce to string            
             return self
 
     def shortlabel(self, label=None):
@@ -72,8 +103,8 @@ class Detection(BoundingBox):
             return self._shortlabel if self._shortlabel is not None else self.category()
 
     def label(self, label):
-        """Alias for category"""
-        return self.category(label)
+        """Alias for category to update both category and shortlabel"""
+        return self.category(label, shortlabel=True)
 
     def id(self):
         return self._id
@@ -81,9 +112,16 @@ class Detection(BoundingBox):
     def clone(self):
         return copy.deepcopy(self)
 
-    def confidence(self):
-        return self._confidence
-    
+    def confidence(self, c=None):
+        if c is None:
+            return self._confidence
+        else:
+            self._confidence = c
+            return self
+
+    def hasattribute(self, k):
+        return isinstance(self.attributes, dict) and k in self.attributes
+
     
 class Track(object):
     """vipy.object.Track class
@@ -124,22 +162,36 @@ class Track(object):
         assert boundary in set(['extend', 'strict']), "Invalid interpolation boundary - Must be ['extend', 'strict']"
         assert interpolation in set(['linear']), "Invalid interpolation - Must be ['linear']"
                 
-        self._id = uuid.uuid1().hex if trackid is None else trackid
+        self._id = uuid.uuid4().hex if trackid is None else trackid
         self._label = category if category is not None else label
         self._shortlabel = self._label if shortlabel is None else shortlabel
         self._framerate = framerate
         self._interpolation = interpolation
-        self._boundary = boundary
+        self._boundary = self.boundary(boundary)
         self.attributes = attributes if attributes is not None else {}        
-        self._keyframes = keyframes
+        self._keyframes = [int(np.round(f)) for f in keyframes]  # coerce to int
         self._keyboxes = boxes
         
         # Sorted increasing frame order
-        if len(keyframes) > 0 and len(boxes) > 0:
+        if len(keyframes) > 0 and len(boxes) > 0 and not all([keyframes[i-1] <= keyframes[i] for i in range(1,len(keyframes))]):
             (keyframes, boxes) = zip(*sorted([(f,bb) for (f,bb) in zip(keyframes, boxes)], key=lambda x: x[0]))
-            self._keyframes = [int(np.round(f)) for f in keyframes]  # coerce to int
+            self._keyframes = list(keyframes)
             self._keyboxes = list(boxes)
-        
+
+    @classmethod
+    def from_json(cls, s):
+        d = json.loads(s) if not isinstance(s, dict) else s
+        return cls(keyframes=[int(f) for f in d['_keyframes']],
+                   boxes=[BoundingBox.from_json(bbs) for bbs in d['_keyboxes']],
+                   category=d['_label'],
+                   confidence=None,
+                   framerate=d['_framerate'],
+                   interpolation=d['_interpolation'],
+                   boundary=d['_boundary'],
+                   shortlabel=d['_shortlabel'],
+                   attributes=d['attributes'],
+                   trackid=d['_id'])
+                   
     def __repr__(self):
         strlist = []
         if self.category() is not None:
@@ -171,43 +223,47 @@ class Track(object):
                     sorted(self.keyframes()) == list(self.keyframes()))
     
     def dict(self):
-        return {'id':self._id, 'label':self.category(), 'shortlabel':self.shortlabel(), 'keyframes':self._keyframes, 'framerate':self._framerate, 
-                'boundingbox':[bb.dict() for bb in self._keyboxes], 'attributes':self.attributes}
+        """Return a python dictionary containing the relevant serialized attributes suitable for JSON encoding"""
+        return self.json(encode=False)
 
-    def add(self, keyframe, box):
-        """Add a new keyframe and associated box to track, preserve sorted order of keyframes"""
-        assert isinstance(box, BoundingBox), "Invalid input - Box must be vipy.geometry.BoundingBox()"
-        assert box.isvalid(), "Invalid input - Box must be non-degenerate"
-        self._keyframes.append(keyframe)
-        self._keyboxes.append(box)
+    def json(self, encode=True):
+        d = {k:v if k != '_keyboxes' else [bb.json(encode=False) for bb in v] for (k,v) in self.__dict__.items()}
+        d['_keyframes'] = [int(f) for f in self._keyframes]
+        return json.dumps(d) if encode else d
+    
+    def add(self, keyframe, bbox, strict=True):
+        """Add a new keyframe and associated box to track, preserve sorted order of keyframes. 
+
+           -strict [bool]:  If box is degenerate, throw an exception if strict=True, otherwise just don't add it
+        """
+        assert isinstance(bbox, BoundingBox), "Invalid input - Box must be vipy.geometry.BoundingBox()"
+        assert strict is False or bbox.isvalid(), "Invalid input - Box must be non-degenerate"
+        if not bbox.isvalid():
+            return self  # just don't add it 
+        self._keyframes.append(int(keyframe))
+        self._keyboxes.append(bbox)
         if len(self._keyframes) > 1 and keyframe < self._keyframes[-2]:
             (self._keyframes, self._keyboxes) = zip(*sorted([(f,bb) for (f,bb) in zip(self._keyframes, self._keyboxes)], key=lambda x: x[0]))        
             self._keyframes = list(self._keyframes)
             self._keyboxes = list(self._keyboxes)
         return self
 
-    def update(self, keyframe, box):
+    def update(self, keyframe, bbox):
         if keyframe in self._keyframes:
             self.delete(keyframe)
-        self.add(keyframe, box)
+        self.add(keyframe, bbox)
         return self
         
     def replace(self, keyframe, box):
-        """Replace a keyframe and associated box to track, preserve sorted order of keyframes"""
-        assert isinstance(box, BoundingBox), "Invalid input - Box must be vipy.geometry.BoundingBox()"
-        assert box.isvalid(), "Invalid input - Box must be non-degenerate"
-        assert keyframe in self._keyframes, "Keyframe not found"
-        self._keyboxes[self._keyframes.index(keyframe)] = box
-        return self
+        """Replace the keyframe and associated box(es), preserve sorted order of keyframes"""
+        return self.delete(keyframe).add(keyframe, box)
 
     def delete(self, keyframe):
         """Replace a keyframe and associated box to track, preserve sorted order of keyframes"""
-        assert keyframe in self._keyframes, "Keyframe not found"
-        k = self._keyframes.index(keyframe)
-        del self._keyboxes[k]
-        del self._keyframes[k]
-        if len(self._keyframes) == 0:
-            warnings.warn('[vipy.object.Track]: Attempting to delete keyframe %d from empty track "%s" - Ignoring' % (keyframe, str(self)))
+        while keyframe in self._keyframes:
+            k = self._keyframes.index(keyframe)
+            del self._keyboxes[k]
+            del self._keyframes[k]
         return self
     
     def keyframes(self):
@@ -227,7 +283,8 @@ class Track(object):
         
     def meanshape(self):
         """Return the mean (width,height) of the box during the track"""
-        return np.mean([bb.shape() for bb in self._keyboxes], axis=0)
+        s = np.mean([bb.shape() for bb in self._keyboxes], axis=0)
+        return (float(s[0]), float(s[1]))
             
     def framerate(self, fps):
         """Resample keyframes from known original framerate set by constructor to be new framerate fps"""
@@ -237,10 +294,10 @@ class Track(object):
         return self
         
     def startframe(self):
-        return np.min(self._keyframes) if len(self._keyframes)>0 else None
+        return int(min(self._keyframes)) if len(self._keyframes)>0 else None
 
     def endframe(self):
-        return np.max(self._keyframes) if len(self._keyframes)>0 else None
+        return int(max(self._keyframes)) if len(self._keyframes)>0 else None
 
     def _linear_interpolation(self, k):
         """Linear bounding box interpolation at frame=k given observed boxes (x,y,w,h) at keyframes.  
@@ -248,32 +305,34 @@ class Track(object):
         If self._boundary='extend', then boxes are repeated if the interpolation is outside the keyframes
         If self._boundary='strict', then interpolation returns None if the interpolation is outside the keyframes
         """
+        assert not self.isempty(), "Degenerate object for interpolation"
         (xmin, ymin, width, height) = zip(*[bb.to_xywh() for bb in self._keyboxes])
-        d = Detection(xmin=np.interp(k, self._keyframes, xmin),
-                      ymin=np.interp(k, self._keyframes, ymin),
-                      width=np.interp(k, self._keyframes, width),
-                      height=np.interp(k, self._keyframes, height),
+        d = Detection(xmin=float(np.interp(k, self._keyframes, xmin)),
+                      ymin=float(np.interp(k, self._keyframes, ymin)),
+                      width=float(np.interp(k, self._keyframes, width)),
+                      height=float(np.interp(k, self._keyframes, height)),
                       category=self.category(),
                       shortlabel=self.shortlabel())
         d.attributes['trackid'] = self.id()  # for correspondence of detections to tracks
-        return d if self._boundary == 'extend' else (None if not self.during(k) else d)
+        return d if self._boundary == 'extend' or self.during(k) else None
 
 
-    def category(self, label=None):
+    def category(self, label=None, shortlabel=True):
         if label is not None:
-            self._label = label
+            self._label = str(label)  # coerce to string
+            self._shortlabel = str(label) if shortlabel else self._shortlabel  # coerce to string
             return self
         else:
             return self._label
     
     def label(self, label):
         """Alias for category"""
-        return self.category(label)
+        return self.category(label, shortlabel=True)
         
     def shortlabel(self, label=None):
         """A optional shorter label string to show as a caption in visualizations"""                
         if label is not None:
-            self._shortlabel = label
+            self._shortlabel = str(label)  # coerce to string
             return self
         else:
             return self._shortlabel
@@ -285,7 +344,7 @@ class Track(object):
 
     def offset(self, dt=0, dx=0, dy=0):
         self._keyboxes = [bb.offset(dx, dy) for bb in self._keyboxes]
-        self._keyframes = list(np.array(self._keyframes) + dt)
+        self._keyframes = [int(f+dt) for f in self._keyframes]
         return self
 
     def frameoffset(self, dx, dy):
@@ -293,6 +352,14 @@ class Track(object):
         assert len(self.keyboxes()) == len(dx) and len(self.keyboxes()) == len(dy)
         self._keyboxes = [bb.offset(dx=x, dy=y) for (bb, (x, y)) in zip(self._keyboxes, zip(dx, dy))]
         return self
+
+    def truncate(self, startframe=None, endframe=None):
+        """Truncate a track so that any keyframes less than startframe or greater than or equal to endframe are removed"""
+        keyframes = copy.deepcopy(self.keyframes())
+        for k in keyframes:
+            if ((startframe is not None and k < startframe) or (endframe is not None and k >= endframe)):
+                self.delete(k)
+        return self
         
     def rescale(self, s):
         """Rescale track boxes by scale factor s"""
@@ -349,10 +416,23 @@ class Track(object):
         return copy.deepcopy(self)
 
     def boundingbox(self):
-        """The bounding box of a track is the smallest spatial box that contains all of the detections, or None if there are no detections"""
+        """The bounding box of a track is the smallest spatial box that contains all of the detections within startframe and endframe, or None if there are no detections"""
         d = self._keyboxes[0].clone() if len(self._keyboxes) >= 1 else None
-        return d.union(self._keyboxes[1:]) if (d is not None and len(self._keyboxes) >= 2) else d
+        return d.union([bb for (k,bb) in zip(self._keyframes[1:], self._keyboxes[1:]) if self.during(k)]) if (d is not None and len(self._keyboxes) >= 2) else d
 
+    def pathlength(self):
+        """The path length of a track is the cumulative Euclidean distance in pixels that the box travels"""
+        return float(np.sum([bb_next.dist(bb_prev) for (bb_next, bb_prev) in zip(self._keyboxes[1:], self._keyboxes[0:-1])])) if len(self._keyboxes)>1 else 0.0
+        
+    def boundary(self, b=None):
+        if b is None:
+            return self._boundary
+        else:
+            assert b in ['strict', 'extend']
+            self._boundary = b
+            return self
+    
+    
     def clip(self, startframe, endframe):
         """Clip a track to be within (startframe,endframe) with strict boundary handling"""
         if self[startframe] is not None:
@@ -381,14 +461,14 @@ class Track(object):
         assert isinstance(other, Track), "invalid input - Must be vipy.object.Track()"
         startframe = max(self.startframe(), other.startframe())
         endframe = min(self.endframe(), other.endframe())
-        return np.mean([self[startframe].iou(other[startframe]), self[endframe].iou(other[endframe])]) if endframe >= startframe else 0.0        
+        return float(np.mean([self[startframe].iou(other[startframe]), self[endframe].iou(other[endframe])]) if endframe >= startframe else 0.0)
 
     def segmentiou(self, other, dt=5):
         """Compute the mean spatial IoU between two tracks at the overlapping segment, sampling by dt.  Useful for track continuation for densely overlapping tracks"""
         assert isinstance(other, Track), "invalid input - Must be vipy.object.Track()"
         startframe = max(self.startframe(), other.startframe())
         endframe = min(self.endframe(), other.endframe())   # inclusive
-        return np.mean([self[min(k,endframe)].iou(other[min(k,endframe)]) for k in range(startframe, endframe, dt)]) if endframe >= startframe else 0.0 
+        return float(np.mean([self[min(k,endframe)].iou(other[min(k,endframe)]) for k in range(startframe, endframe, dt)]) if endframe >= startframe else 0.0)
         
     def rankiou(self, other, rank, dt=1, n=None):
         """Compute the mean spatial IoU between two tracks per frame in the range (self.startframe(), self.endframe()) using only the top-k (rank) frame overlaps
@@ -400,7 +480,7 @@ class Track(object):
         assert dt >= 1
         dt = max(1, int(len(self)/n) if n is not None else dt)
         frames = [self.startframe()] + list(range(self.startframe()+dt, self.endframe(), dt)) + [self.endframe()]
-        return np.mean(sorted([self[k].iou(other[k]) if (self.during(k) and other.during(k)) else 0.0 for k in frames])[-rank:])
+        return float(np.mean(sorted([self[k].iou(other[k]) if (self.during(k) and other.during(k)) else 0.0 for k in frames])[-rank:]))
 
     def percentileiou(self, other, percentile, dt=1, n=None):
         """Percentile iou returns rankiou for rank=percentile*len(self)"""
@@ -436,7 +516,7 @@ class Track(object):
         self._keyboxes = [bb.clone().medianshape(bbnbrs) for (bb, bbnbrs) in zip(self._keyboxes, chunklistwithoverlap(self._keyboxes, width, width-1))]
         return self
 
-    def spline(self, smoothingfactor=None):
+    def spline(self, smoothingfactor=None, strict=True):
         """Track smoothing by cubic spline fit, will return resampled dt=1 track.  Smoothing factor will increase with smoothing > 1 and decrease with 0 < smoothing < 1
         
            This function requires optional package scipy
@@ -444,15 +524,45 @@ class Track(object):
         try_import('scipy', 'scipy');  import scipy.interpolate;
         assert smoothingfactor is None or smoothingfactor > 0
         t = self.clone().resample(dt=1)
-        s = smoothingfactor * len(self._keyframes) if smoothingfactor is not None else None
-        (xmin, ymin, xmax, ymax) = zip(*[bb.to_ulbr() for bb in t._keyboxes])
-        f_xmin = scipy.interpolate.UnivariateSpline(t._keyframes, xmin, check_finite=False, s=s)
-        f_ymin = scipy.interpolate.UnivariateSpline(t._keyframes, ymin, check_finite=False, s=s)
-        f_xmax = scipy.interpolate.UnivariateSpline(t._keyframes, xmax, check_finite=False, s=s)
-        f_ymax = scipy.interpolate.UnivariateSpline(t._keyframes, ymax, check_finite=False, s=s)
-        (self._keyframes, self._keyboxes) = zip(*[(k, BoundingBox(xmin=float(f_xmin(k)), ymin=float(f_ymin(k)), xmax=float(f_xmax(k)), ymax=float(f_ymax(k)))) for k in range(self.startframe(), self.endframe())])
+        try:
+            assert len(t._keyframes) > 4, "Invalid length for spline interpolation"        
+            s = smoothingfactor * len(self._keyframes) if smoothingfactor is not None else None
+            (xmin, ymin, xmax, ymax) = zip(*[bb.to_ulbr() for bb in t._keyboxes])
+            f_xmin = scipy.interpolate.UnivariateSpline(t._keyframes, xmin, check_finite=False, s=s)
+            f_ymin = scipy.interpolate.UnivariateSpline(t._keyframes, ymin, check_finite=False, s=s)
+            f_xmax = scipy.interpolate.UnivariateSpline(t._keyframes, xmax, check_finite=False, s=s)
+            f_ymax = scipy.interpolate.UnivariateSpline(t._keyframes, ymax, check_finite=False, s=s)
+            (self._keyframes, self._keyboxes) = zip(*[(k, BoundingBox(xmin=float(f_xmin(k)), ymin=float(f_ymin(k)), xmax=float(f_xmax(k)), ymax=float(f_ymax(k)))) for k in range(self.startframe(), self.endframe())])
+        except Exception as e:
+            if not strict:
+                print('[vipy.object.track]: spline smoothing failed with error "%s" - Returning unsmoothed track' % (str(e)))
+                return self
+            else:
+                raise
         return self
 
+    def linear_extrapolation(self, k, smoothingfactor=None):
+        """Track extrapolation by linear fit.  Smoothing factor will increase with smoothing > 1 and decrease with 0 < smoothing < 1.
+        
+           Requires at least 3 keyboxes.
+           Returned boxes may be degenerate if extrapolation failed. 
+           Requires optional package scipy.
+        """
+        if self.during(k):
+            return self[k]
+        if len(self.keyboxes()) < 3:
+            return self.nearest_keybox(k)
+        try_import('scipy', 'scipy');  import scipy.interpolate;
+        assert smoothingfactor is None or smoothingfactor > 0
+        assert len(self._keyframes) >= 3, "Invalid length for linear extrapolation"
+        s = smoothingfactor * len(self._keyframes) if smoothingfactor is not None else None
+        (xmin, ymin, width, height) = zip(*[bb.to_xywh() for bb in self._keyboxes])
+        f_xmin = scipy.interpolate.UnivariateSpline(self._keyframes, xmin, check_finite=False, s=s, k=1)
+        f_ymin = scipy.interpolate.UnivariateSpline(self._keyframes, ymin, check_finite=False, s=s, k=1)
+        f_width = scipy.interpolate.UnivariateSpline(self._keyframes, width, check_finite=False, s=s, k=1)
+        f_height = scipy.interpolate.UnivariateSpline(self._keyframes, height, check_finite=False, s=s, k=1)
+        return Detection(xmin=float(f_xmin(k)), ymin=float(f_ymin(k)), width=float(f_width(k)), height=float(f_height(k)))
+    
     def imclip(self, width, height):
         """Clip the track to the image rectangle (width, height).  If a keybox is outside the image rectangle, remove it otherwise clip to the image rectangle. 
            This operation can change the length of the track and the size of the keyboxes.  The result may be an empty track if the track is completely outside
@@ -474,7 +584,34 @@ class Track(object):
         (self._keyboxes, self._keyframes) = (list(self._keyboxes), list(self._keyframes))
         return self
 
+    def significant_digits(self, n):
+        self._keyboxes = [bb.significant_digits(n) for bb in self._keyboxes]
+        return self
 
+    def velocity(self, f, dt=5):
+        """Return the (x,y) track velocity at frame f in units of pixels per frame computed by finite difference"""
+        assert f >= 0 and dt > 0 and self.during(f)              
+        return (self.velocity_x(f, dt), self.velocity_y(f, dt))
+
+    def velocity_x(self, f, dt=5):
+        """Return the (x) track velocity at frame f in units of pixels per frame computed by finite difference"""
+        assert f >= 0 and dt > 0 and self.during(f)
+        return (self[f].centroid_x() - self[f-dt].centroid_x())/float(dt)
+
+    def velocity_y(self, f, dt=5):
+        """Return the (y) track velocity at frame f in units of pixels per frame computed by finite difference"""
+        assert f >= 0 and dt > 0 and self.during(f)
+        return (self[f].centroid_y() - self[f-dt].centroid_y())/float(dt)
+    
+    def nearest_keyframe(self, f):
+        """Nearest keyframe to frame f"""
+        return self._keyframes[int(np.abs(np.array(self._keyframes) - f).argmin())]
+
+    def nearest_keybox(self, f):
+        """Nearest keybox to frame f"""
+        return self._keyboxes[int(np.abs(np.array(self._keyframes) - f).argmin())]
+    
+    
 def non_maximum_suppression(detlist, conf, iou, bycategory=False):
     """Compute non-maximum suppression of a list of vipy.object.Detection() based on spatial IOU threshold (iou) and a confidence threshold (conf)"""
     assert all([isinstance(d, Detection) for d in detlist])
