@@ -8,11 +8,12 @@ import numpy as np
 import tempfile
 import time
 from time import gmtime, strftime, localtime
+from datetime import datetime
+import pytz
 import sys
 import csv
 import hashlib
 import shutil
-import json
 import shelve
 import re
 import uuid
@@ -28,6 +29,19 @@ import socket
 import warnings
 import copy
 
+try:
+    import ujson as json  # faster
+except ImportError:
+    import json
+
+
+
+def mergedict(d1, d2):
+    assert isinstance(d1, dict) and isinstance(d2, dict)
+    d = copy.deepcopy(d1)
+    d.update(d2)
+    return d
+
 
 def hascache():
     """Is the VIPY_CACHE environment variable set?"""
@@ -47,7 +61,7 @@ def try_import(package, pipname=None, message=None):
         if message is not None:
             raise ImportError(message)
         else:
-            raise ImportError('Optional package "%s" not installed -  Run "pip install %s" ' % (package, package if pipname is None else pipname))
+            raise ImportError('Optional package "%s" not installed -  Run "pip install %s" or "pip install vipy[all]" ' % (package, package if pipname is None else pipname))
 
 
 def findyaml(basedir):
@@ -255,7 +269,6 @@ def imwritejet(img, imfile=None):
 def isuint8(img):
     return isnumpy(img) and img.dtype == np.dtype('uint8')
 
-
 def isnumber(x):
     """Is the input a python type of a number or a string containing a number?"""
     return isinstance(x, (int, float)) or (isnumpy(x) and np.isscalar(x)) or (isstring(x) and isfloat(x))
@@ -394,117 +407,191 @@ def loadmat73(matfile, keys=None):
         return np.array(f)
 
 
-def saveas(vars, outfile=None, type='dill'):
+def _json_class_registry():
+    import vipy.video
+    import vipy.image
+    registry = {"<class 'vipy.video.Scene'>":vipy.video.Scene.from_json,
+                "<class 'vipy.video.Video'>":vipy.video.Video.from_json,
+                "<class 'vipy.video.VideoCategory'>":vipy.video.VideoCategory.from_json,
+                "<class 'vipy.image.Image'>":vipy.image.Image.from_json,
+                "<class 'vipy.image.ImageCategory'>":vipy.image.ImageCategory.from_json,
+                "<class 'vipy.image.ImageDetection'>":vipy.image.ImageDetection.from_json,            
+                "<class 'vipy.image.Scene'>":vipy.image.Scene.from_json,
+                "<class 'vipy.geometry.BoundingBox'>":vipy.geometry.BoundingBox.from_json,
+                "<class 'vipy.object.Track'>":vipy.object.Track.from_json,
+                "<class 'vipy.object.Detection'>":vipy.object.Detection.from_json,
+                "<class 'vipy.activity.Activity'>":vipy.activity.Activity.from_json}
+    try:
+        import pycollector.video
+        registry.update( {"<class 'pycollector.video.Video'>":pycollector.video.Video.from_json} )
+    except:
+        registry.update( {"<class 'pycollector.video.Video'>":lambda x: exec("raise ValueError(\"<class 'pycollector.video.Video'> not found - Run 'pip install pycollector' \")")})        
+    try:
+        import pycollector.admin.video
+        registry.update( {"<class 'pycollector.admin.video.Video'>":pycollector.admin.video.Video.from_json} )
+    except:
+        registry.update( {"<class 'pycollector.admin.video.Video'>":lambda x: exec("raise ValueError(\"<class 'pycollector.admin.video.Video'> not found - This is for admin use only \")")})        
+
+    return registry
+            
+
+def saveas(vars, outfile=None, format='dill'):
     """Save variables as a dill pickled file"""
     outfile = temppickle() if outfile is None else outfile
-    if type in ['dill']:
+
+    remkdir(filepath(outfile))
+    if format in ['dill']:
         dill.dump(vars, open(outfile, 'wb'))
         return outfile
+    
+    elif format == 'json':
+        saveobj = vars
+        class_registry = _json_class_registry()
+        if isinstance(saveobj, list) and all([str(type(d)) in class_registry for d in saveobj]):
+            j = [{str(type(d)):d.json(encode=False)} for d in saveobj] if isinstance(saveobj, list) else ({str(type(d)):d.json(encode=False)} for d in saveobj)
+        elif str(type(saveobj)) in class_registry:
+            j = {str(type(saveobj)):saveobj.json(encode=False)}
+        else:
+            j = saveobj
+
+        s = json.dumps(j, ensure_ascii=False)  # load to memory (faster than json.dump), will throw exception if it cannot serialize
+        with open(outfile, 'w') as f:
+            f.write(s)            
+        return outfile
+        
     else:
-        raise ValueError('unknown serialization type "%s"' % type)
+        raise ValueError('unknown serialization format "%s"' % format)
 
     return outfile
 
 
-def loadas(infile, type='dill'):
+def loadas(infile, format='dill'):
     """Load variables from a dill pickled file"""
-    if type in ['dill']:
-        try:
-            return dill.load(open(infile, 'rb'))
-        except Exception:
-            import pickle
-            with open(infile, 'rb') as pfp:
-                fv = pickle.load(pfp, encoding="bytes")
-            return fv
+    
+    if format in ['dill', 'pkl']:
+        return dill.load(open(infile, 'rb'))
+    elif format == 'json':
+        with open(infile, 'r') as f:
+            loadobj = json.load(f)
+        class_registry = _json_class_registry()
+        assert isinstance(loadobj, list) or isinstance(loadobj, dict), "invalid vipy JSON serialization format"
+        if isinstance(loadobj, list) and all([isinstance(d, dict) for d in loadobj]) and all([c in class_registry for d in loadobj for (c,v) in d.items()]):
+            return [class_registry[c](v) for d in loadobj for (c,v) in d.items()]
+        elif isinstance(loadobj, dict) and all([c in class_registry for (c,d) in loadobj.items()]):
+            obj = [class_registry[c](v) for (c,v) in loadobj.items()]
+            return obj[0] if len(obj) == 1 else obj
+        else:
+            return loadobj
     else:
-        raise ValueError('unknown serialization type "%s"' % type)
+        raise ValueError('unknown serialization format "%s"' % format)
 
 
 def load(infile):
-    """Load variables from a relocatable dill pickled file
+    """Load variables from a relocatable archive file format, either Dill Pickle or JSON.
        
        Loading is performed by attemping the following:
 
-       1. load the pickle file
-       2. If the loaded object is a vipy object (or iterable) and the relocatable path /$PATH is present, try to repath it to the directory containing this pickle
+       1. load the pickle or json file
+       2. If the loaded object is a vipy object (or iterable) and the relocatable path /$PATH is present, try to repath it to the directory containing this archive
        3. If the resulting files are not found, throw a warning
 
     """
-    obj = loadas(infile, type='dill')
-    testobj = copy.deepcopy(tolist(obj)[0])
+    infile = os.path.abspath(os.path.expanduser(infile))
+    if ispkl(infile):
+        format = 'dill'
+    elif isjsonfile(infile):
+        format = 'json'
+    else:
+        raise ValueError('unknown file type')
+    
+    obj = loadas(infile, format=format)
+    try:
+        testobj = tolist(obj)[0]
+    except:
+        raise ValueError('Invalid archive "%s"' % infile)
+
+    # Relocatable object?
     if hasattr(testobj, 'filename') and testobj.filename() is not None and '/$PATH' in testobj.filename():
-        testobj = repath(testobj, '/$PATH', filepath(os.path.abspath(infile)))  # attempt to rehome /$PATH/to/me.jpg -> /NEWPATH/to/me.jpg where NEWPATH=filepath(infile)
+        testobj = repath(copy.deepcopy(testobj), '/$PATH', filepath(os.path.abspath(infile)))  # attempt to rehome /$PATH/to/me.jpg -> /NEWPATH/to/me.jpg where NEWPATH=filepath(infile)
         if os.path.exists(testobj.filename()):       # file found
             obj = repath(obj, '/$PATH', filepath(os.path.abspath(infile)))      # rehome everything to the same root path as the pklfile
         else:
             warnings.warn('Loading "%s" that contains redistributable paths - Use vipy.util.distload("%s", datapath="/path/to/your/data") to rehome absolute file paths' % (infile, infile))
-    elif hasattr(testobj, 'hasfilename') and testobj.filename() is not None and not testobj.hasfilename(): 
-        warnings.warn('Loading "%s" that contains absolute filepaths - The relocated filename "%s" does not exist' % (infile, testobj.filename()))
+    elif hasattr(testobj, 'filename') and testobj.filename() is not None and not os.path.exists(testobj.filename()):
+        if hasattr(testobj, 'hasurl') and testobj.hasurl():
+            warnings.warn('Invalid file - Loading "%s" that contains filename "%s" which does not exist - This must be downloaded from the provided url() using download()' % (infile, testobj.filename()))
+        elif not os.path.isabs(testobj.filename()):
+            warnings.warn('Invalid file - Loading "%s" that contains relative path "%s" which does not exist - Loading archives with relative paths must be loaded from the same directory containing the videos' % (infile, testobj.filename()))
+        else:
+            warnings.warn('Invalid file - Loading "%s" that contains absolute path "%s" which does not exist' % (infile, testobj.filename()))
     return obj
 
 
+def save(vars, outfile, mode=None):
+    """Save variables to an archive file"""
+    allowable = set(['.pkl', '.json'])
+    if ispkl(outfile):
+        format = 'dill'
+    elif isjsonfile(outfile):
+        format = 'json'
+    else:
+        raise ValueError('Unknown file extension "%s" - must be in %s' % (fileext(outfile), str(allowable)))
+    
+    outfile = saveas(vars, outfile, format=format)
+    if mode is not None:
+        chmod(outfile, mode)
+    return outfile
+
+
 def distload(infile, datapath, srcpath='/$PATH'):
-    """Load a redistributable pickle file that replaces absolute paths in datapath with srcpath.  See also vipy.util.distsave()"""
+    """Load a redistributable pickle file that replaces absolute paths in datapath with srcpath.  See also vipy.util.distsave(),
+       This function has been deprecated, all archives should be distributed with relative paths
+    """
     return repath(load(infile), srcpath, datapath)
 
 
+def distsave(vars, datapath, outfile=None, mode=None, dstpath='/$PATH'):
+    """Save a archive file for redistribution, where datapath is replaced by dstpath.  Useful for redistribuing pickle files with absolute paths.  See also vipy.util.distload().
+       This function has been deprecated, all archives should be distributed with relative paths       
+    """
+    vars = vars if (datapath is None or not isvipyobject(vars)) else repath(vars, datapath, dstpath) 
+    return save(vars, outfile, mode)
+
+
 def repath(v, srcpath, dstpath):
     """Change the filename with prefix srcpath to dstpath, for any element in v that supports the filename() api"""
     if not islist(v) and (hasattr(v, 'filename') and hasattr(v, 'clone')):
         vc = v.filename( v.filename().replace(os.path.normpath(srcpath), os.path.normpath(dstpath))) if v.filename() is not None else v
     elif islist(v) and all([(hasattr(vv, 'filename') and hasattr(vv, 'clone')) for vv in v]):
         vc = [vv.filename( vv.filename().replace(os.path.normpath(srcpath), os.path.normpath(dstpath))) if vv.filename() is not None else vv for vv in v ]
+    elif isstring(v):
+        vc = v.replace(os.path.normpath(srcpath), os.path.normpath(dstpath))
     else:
         raise ValueError('Input must be a singleton or list of vipy.image.Image() or vipy.video.Video() objects, not type "%s"' % (str(type(v))))
     return vc
     
 
-def scpsave(v):
+def scpsave(V):
     import vipy.image
     import vipy.video
-    if (isinstance(v, vipy.image.Image) or isinstance(v, vipy.video.Video)) and v.hasfilename():        
-        vc = v.clone().url('scp://%s:%s' % (socket.gethostname(), v.filename())).nofilename()
-        v = vc
-    elif islist(v) and all([isinstance(vv, vipy.image.Image) or isinstance(vv, vipy.video.Video) for vv in v]):
-        vc = [vv.clone().url('scp://%s:%s' % (socket.gethostname(), vv.filename())).nofilename() for vv in v]
-        v = vc
-    return 'scp://%s:%s' % (socket.gethostname(), save(v))
+    if (isinstance(V, vipy.image.Image) or isinstance(V, vipy.video.Video)) and v.hasfilename():        
+        v = V
+        v = v.clone().url('scp://%s:%s' % (socket.gethostname(), v.filename())).nofilename()
+    elif islist(V) and all([isinstance(v, vipy.image.Image) or isinstance(v, vipy.video.Video) for v in V]):
+        v = [v.clone().url('scp://%s:%s' % (socket.gethostname(), v.abspath().filename())).nofilename() for v in V]
+    else:
+        pass  # no vipy objects
+
+    pklfile = 'scp://%s:%s' % (socket.gethostname(), save(v, temppkl()))
+    cmd = "V = vipy.util.scpload('%s')" % pklfile
+    print('[vipy.util.scpsave]: On a remote machine where you have public key ssh access to this machine run:\n>>> %s\n' % cmd)
+    return pklfile
 
 
 def scpload(url):
     import vipy.downloader
     return load(vipy.downloader.scp(url, templike(url)))
 
-                
-def save(vars, outfile=None, mode=None):
-    """Save variables as a dill pickled file"""
-    outfile = temppickle() if outfile is None else outfile
-    outfile = saveas(vars, outfile, type='dill')
-    if mode is not None:
-        chmod(outfile, mode)
-    return outfile
-
-
-def distsave(vars, datapath, outfile=None, mode=None, dstpath='/$PATH'):
-    """Save a pickle file for redistribution, where datapath is replaced by dstpath.  Useful for redistribuing pickle files with absolute paths.  See also vipy.util.distload()"""
-    vars = vars if (datapath is None or not isvipyobject(vars)) else repath(vars, datapath, dstpath) 
-    return save(vars, outfile, mode)
-
-
-def fastsave(v, outfile=None):
-    """Save variables as a cPickle file with the highest protocol - useful
-    for very large custom classes"""
-    with open(outfile, 'wb') as fp:
-        cPickle.dump(v, fp, cPickle.HIGHEST_PROTOCOL)
-    return outfile
-
-
-def fastload(infile):
-    """Load file saved with fastsave"""
-    with open(infile, 'rb') as fp:
-        v = cPickle.load(fp)
-    return v
-
 
 def load_opencv_yaml(yamlfile):
     """Load a numpy array from YAML file exported from OpenCV"""
@@ -598,19 +685,13 @@ def ispkl(filename):
     return filename[-4:] == '.pkl' if isstring(filename) and \
         len(filename) >= 4 else False
 
+def ishtml(filename):
+    """Is the file an HTMLfile"""
+    return fileext(filename).lower() == '.html'
 
 def ispickle(filename):
     """Is the file a pickle archive file"""
-    return isfile(filename) and os.path.exists(filename) and \
-        fileext(filename).lower() in ['.pk', '.pkl']
-
-
-def ndmax(A):
-    return np.unravel_index(A.argmax(), A.shape)
-
-
-def ndmin(A):
-    return np.unravel_index(A.argmin(), A.shape)
+    return isfile(filename) and os.path.exists(filename) and (fileext(filename) is not None) and fileext(filename).lower() in ['.pk', '.pkl']
 
 
 def ishdf5(path):
@@ -631,9 +712,11 @@ def filebase(filename):
     return base
 
 
-def filepath(filename):
-    """Return /a/b for filename /a/b/c.ext"""
+def filepath(filename, depth=0):
+    """Return /a/b/c for filename /a/b/c/d.ext, /a/b for filename /a/b/c/d.ext if depth=1, etc"""
     (head, tail) = os.path.split(filename)
+    for k in range(depth):
+        (head, tail) = os.path.split(head)           
     return head
 
 
@@ -695,6 +778,11 @@ def dirlist(indir):
                 not is_hiddenfile(item))]
 
 
+def dirlist_sorted_bycreation(indir):
+    """Sort the directory list from newest first to oldest last by creation date"""
+    return sorted(dirlist(indir), key=lambda d: os.stat(d).st_ctime, reverse=True)
+
+
 def extlist(indir, ext):
     """return list of files with absolute path in a directory that have
     the provided extension (with the prepended dot, ext='.mp4')"""
@@ -703,6 +791,9 @@ def extlist(indir, ext):
             if fileext(item) is not None
             and (fileext(item).lower() == ext.lower())]
 
+def listext(indir, ext):
+    """Alias for extlist"""
+    return extlist(indir, ext)
 
 def jsonlist(indir):
     """return list of fJSON iles with absolute path in a directory"""
@@ -778,10 +869,14 @@ def csvlist(imdir):
             if iscsv(item)]
 
 
-def pklist(imdir):
+def pklist(indir):
     """Return a list of absolute paths of *.pk files in current directory"""
-    return [os.path.join(imdir, item) for item in os.listdir(imdir)
-            if ispickle(os.path.join(imdir, item))]
+    return listpkl(indir)
+
+def listpkl(indir):
+    """Return a list of absolute paths of *.pk files in current directory"""
+    return [os.path.join(indir, item) for item in os.listdir(indir)
+            if ispickle(os.path.join(indir, item))]
 
 
 def txtlist(imdir):
@@ -899,6 +994,8 @@ def tolist(x):
         return x
     elif isinstance(x, tuple):        
         return list(x)
+    elif isinstance(x, set):        
+        return list(x)
     else:
         return [x]
 
@@ -1045,6 +1142,11 @@ def isarchive(filename):
         else:
             return False
 
+def istgz(filename):
+    return filename[-4:] == '.tgz' or filename[-7:] == '.tar.gz'
+
+def isbz2(filename):
+    return filename[-4:] == '.bz2' or filename[-8:] == '.tar.bz2'
 
 def tempfilename(suffix):
     """Create a temporary filename $TEMPDIR/$UUID.suffix, suffix should include the dot such as suffix='.jpg', """
@@ -1237,6 +1339,9 @@ def timestamp():
     """Return date and time string in form DDMMMYY_HHMMSS"""
     return str.upper(strftime("%d%b%y_%I%M%S%p", localtime()))
 
+def clockstamp():
+    """Datetime stamp in eastern timezone with second resolution"""    
+    return datetime.now().astimezone(pytz.timezone("US/Eastern")).strftime("%Y-%m-%dT%H:%M:%S%z")    
 
 def minutestamp():
     """Return date and time string in form DDMMMYY_HHMM"""
